"""
Workspace connection for dask-yarn

Dask-yarn does not really benefit from any sql magics. It's included here
due to being a common analytic pattern for distributed analytics using 
yarn.
"""


class Daskyarn:

    def __init__(self,kwargs,command=None):
        self.conf=kwargs
        if command is not None:
            self.command=command

    def init(self):
        from dask_yarn import YarnCluster
        from dask.distributed import Client
        if self._client is None:
            self._client = Client(YarnCluster(**self.conf))
        return(self._client)
    
    def conn(self):
        return(None)

    def qry(self,qry,r_type="df",limit=30):   
        return(None)



"""
Data Science Work Space - Duct

Duct provides a single class oject that is intended to consolidate
all analytical source data connection patterns.
'duct' was deliberately chosen to consolidate the collection of connection
components, but avoid names like pipe, connection, stream to avoid naming
confusion as well as to not limit the types of work space components included

To use duct:

```python
from dsws.duct import Duct
duct_conf = {'hive': {'class': 'ws_impyla',
                      'magic_handle': 'hive',
                      'kwargs': {'host': 'node01.ratiocinate.com',
                                 'port': 10000,
                                 'auth_mechanism': 'PLAIN'},
                      'command': 'beeline -u "jdbc:hive2://node01.ratiocinate.com:10000"'}
duct = Duct(duct_conf)
```
"""

import os
from os                 import environ
from pydoc              import locate
from IPython.core.magic import register_line_cell_magic
from dsws.util          import sp
from IPython.display    import HTML
from dsws.util          import launch_term

def get_cls(ws_str):
    """
    get_cls - helper function to return the primary class
    object within a dsws submodule. Requires two nameing 
    conventions to be true of the submodule:
     - The module is lower_case with the second '_' split
     and after matching the name of the class
     - The class name is capitalized
    """
    cap="_".join(ws_str.split("_")[1:]).capitalize()
    return locate('.'.join(['dsws',ws_str,cap]))

def register_magic(handle,cls):
    """
    register_magic - helper function that standardizes
    classes registered magics with the follwoing 
    conventions:
     - An unargumented magic launches a cli session
     - An argumented magic accepts sql and returns
     a display of results. Intended for very concise
     iterations or summary results.
    """
    @register_line_cell_magic(handle)
    def tmp(line='', cell=''):
        qry=str('\n'.join([line,cell]).strip())
        if qry=='':
            launch_term(cls.command)
        else: 
            cls.qry(qry, "disp")
    del tmp
    
class Duct:
    """
    Duct - A class that holds workspaces defined within duct_conf.

    During instantiation the wrokspaces are defined as well as 
    magics are created according to dsws conventions:
     - An unargumented magic launches a cli session
     - An argumented magic accepts sql and returns
     a display of results. Intended for very concise
     iterations or summary results.
    """
    def __init__(self,duct_conf={}):
        for k,v in duct_conf.items():
            kwargs={j:w for j,w in v.items() if j in ('kwargs','command')}
            setattr(self, k, get_cls(v["class"])(**kwargs))
            if 'magic_handle' in v.keys():
                register_magic(v['magic_handle'],getattr(self, k))



"""
Workspace connection for Impyla

ws_impyla is a dsws conn type library that follows PEP-249.
It contains a single class, Impyla.
It is intended to be the primary connection type for hive
and impala connections although other ws connections
may be used for hive or impala via odbc / jdbc.
"""

import pandas          as pd
from impala.dbapi  import connect
from dsws.util     import pretty
from dsws.util     import sp
from dsws.util     import standard_conn_qry
from dsws.util     import no_return
from os import environ as env
from ast           import literal_eval

class Impyla:

    def __init__(self,kwargs,command=None):
        """
        Workspace class from the impyla library

        As convntion for DSWS, the following class methods are
        provided:
         - conn: returns a connection class impala.hiveserver2.HiveServer2Connection
         - qry: precesses a query, return type is dependent upon r_type requested:
            - df/raw: pandas dataframe
            - disp: pretty html form of pandas dataframe
            - msg: experimental - retuens query profile if using impala
            - cmd: standard query form to be processed. Helpful for debugging and logging
        
        Configuring an Impyla class is typically done through dsws.duct, but can also be
        evaluated as:
        ```python
        from dsws.ws_impyla import Impyla
        kwargs = {'host': 'node01.ratiocinate.com',
                  'port': 10000,
                  'auth_mechanism': 'PLAIN'}
        hive = Impyla(kwargs)
        ```
        
        To know which values can be accepted in kwargs run:
        ```python
        from impala.dbapi import connect
        ?connect
        ```
        
        Any values in kwargs that are capitalized and not connection arguments
        will be processed as query set statements.
        """
        self.qryconf={k:v for k,v in kwargs.items() if k.isupper()}
        self.conf={a:kwargs[a] for a in connect.__code__.co_varnames if a in kwargs.keys()}
        if command is not None:
            self.command=command
    
    def conn(self):
        return (connect(**self.conf))

    def qry(self,qry,r_type="df",limit=30):
        r_type = 'df' if r_type=='raw' else r_type
        qry=standard_conn_qry(qry)
        if r_type=="cmd":
            return(qry)
        conn   = self.conn()
        cursor = conn.cursor()
        for k in self.qryconf:
          cursor.execute("SET %s=%s" % (k,self.qryconf[k]))
        for q in qry[:-1]:
            cursor.execute(q)
        if r_type=="disp" and "LIMIT" not in qry[-1].split()[-2].upper() and "SELECT" in qry[-1].split()[1].upper():
            qry[-1] = qry[-1] + " LIMIT " + str(limit)
        cursor.execute(qry[-1])
        if no_return(qry[-1]):
            return(None)
        if r_type in ("df","disp"):
            rslt = pd.read_sql(sql=qry[-1],con=conn)
            if r_type=="disp":
                pretty(rslt,col="#5697cb")
                rslt=None
        elif r_type=="msg":
            cursor.execute(qry[-1])
            rslt=(cursor.get_summary(),cursor.get_profile())
        else:
            rslt=None
        conn.close()    
        return(rslt)



"""
Workspace connection for PheonixDB

ws_pheonix is a dsws conn type library that follows PEP-249.
It contains a single class, Pheonix.
Pheonix uses a connection string to connect or it accepts kwargs to generate a conneciton string.
"""

import pandas          as pd
from phoenixdb import connect
from dsws.util     import pretty
from dsws.util     import sp
from dsws.util     import standard_conn_qry
from dsws.util     import no_return
from os import environ as env
from ast           import literal_eval

class Pheonix:

    def __init__(self,kwargs,command=None):
        """
        Workspace class from the Pheonix library

        As conviention for DSWS, the following class methods are
        provided:
         - conn: returns a connection class pyodbc.Connection
         - qry: precesses a query, return type is dependent upon r_type requested:
            - df/raw: pandas dataframe
            - disp: pretty html form of pandas dataframe
            - cmd: standard query form to be processed. Helpful for debugging and logging
        
        Configuring a pheonix class is typically done through dsws.duct, but can also be
        evaluated as:
        ```python
        from dsws.ws_pheonix import Pheonix
        kwargs = {'database_url':'http://localhost:8765/',
                  'autocommit': True}
        phx = Pheonix(kwargs)
        ```
        """
        self.qryconf={k:v for k,v in kwargs.items() if k.isupper()}
        self.conf={k:v for k,v in kwargs.items() if k not in self.qryconf}
        if command is not None:
            self.command=command
    
    def conn(self):
       return(connect(**self.conf))

    def qry(self,qry,r_type="df",limit=30):
        r_type = 'df' if r_type=='raw' else r_type
        qry=standard_conn_qry(qry)
        if r_type=="cmd":
            return(qry)
        conn   = self.conn()
        cursor = conn.cursor()
        for k in self.qryconf:
          cursor.execute("SET %s=%s" % (k,self.qryconf[k]))
        for q in qry[:-1]:
            cursor.execute(q)
        if r_type=="disp" and "LIMIT" not in qry[-1].split()[-2].upper() and "SELECT" in qry[-1].split()[1].upper():
            qry[-1] = qry[-1] + " LIMIT " + str(limit)
        cursor.execute(qry[-1])
        if no_return(qry[-1]):
            return(None)
        if r_type in ("df","disp"):
            rslt = pd.read_sql(sql=qry[-1],con=conn)
            if r_type=="disp":
                pretty(rslt,col="#5697cb")
                rslt=None
        elif r_type=="msg":
            cursor.execute(qry[-1])
            rslt=(cursor.get_summary(),cursor.get_profile())
        else:
            rslt=None
        conn.close()    
        return(rslt)



"""
Workspace connection for SQL Alchemy

ws_sqlalchemy is a dsws conn type library that follows PEP-249.
It contains a single class, Sqlalchemy.
It is intended to be an alternate connection type for
systems if there is user preference.
"""

import pandas          as pd
from sqlalchemy import create_engine
from dsws.util     import pretty
from dsws.util     import sp
from dsws.util     import standard_conn_qry
from dsws.util     import no_return
from os import environ as env
from ast           import literal_eval

class Sqlalchemy:

    def __init__(self,kwargs,command=None):
        """
        Workspace class from the sqlalchemy library

        As convntion for DSWS, the following class methods are
        provided:
         - conn: returns a sqlalchemy connection class
         - qry: precesses a query, return type is dependent upon r_type requested:
            - df/raw: pandas dataframe
            - disp: pretty html form of pandas dataframe
        
        Configuring a Sqlalchemy class is typically done through dsws.duct, but can also be
        evaluated as:
        ```python
        from dsws.ws_sqlalchemy import Sqlalchemy
        kwargs = {'name_or_url': 'mysql+pymysql://user:password@dbname.db.ratiocinate.com:3306/dbname'}
        sa = Sqlalchemy(kwargs)
        ```
        """
        self.conf=kwargs
        if command is not None:
            self.command=command
    
    def conn(self):
        """
        Creates a sqlalchemy connection from a sqlalchemy engine
        """
        engine = create_engine(**self.conf)
        return engine.connect()    

    def qry(self,qry,r_type="df",limit=30):
        r_type = 'df' if r_type=='raw' else r_type
        qry=standard_conn_qry(qry)
        if r_type=="cmd":
            return(qry)
        conn   = self.conn()
        for q in qry[:-1]:
            conn.execute(q)
        if r_type=="disp" and "LIMIT" not in qry[-1].split()[-2].upper() and "SELECT" in qry[-1].split()[1].upper():
            qry[-1] = qry[-1] + " LIMIT " + str(limit)
        if r_type in ("df","disp"):
            rslt = pd.read_sql(sql=qry[-1],con=conn)
            if r_type=="disp":
                pretty(rslt,col="#ff9280")
                rslt=None
        else:
            rslt=None
        conn.close()    
        return(rslt)



"""
Data Science Workspace Utility

Collection of utility functions used throughout the workspace
connection types:
 - pretty: formatted table output for jupyter sessions
 - sp: subprocess call wrapper for OS operations
 - which_ui: centralize method to determine environment
 from only environment variables and sys.platform
 - launch_term: launches a cli tool in a terminal, determines
 environment within the method
 - launch_url: launches a webpage, works only within jupyter
 and CDSW
 - standard_conn_qry: returns the standard qry form for dsws type conn
 - standard_sess_qry: returns the standard qry form for dsws type sess
 - no_return: helper function for when no return data expected
"""

import subprocess
import os
import sys
import os
from IPython.core.display import display
from IPython.core.display import HTML
import numpy as np
import re
import shlex
import time 


def pretty(df, max_lines=50, col="lightgray"):
    df = df[:max_lines].to_html(index=False)
    df = df.replace('<th>','<th style="text-align:center;background-color:%s">' % col)
    display(HTML("<div align=center>" + df + "</div>"))


def sp(cmd_lst,shell=False):
    alt_env = os.environ.copy()
    #REF: Haven't neded yet, can use alt_env
    #alt_env["PATH"] = "/usr/bin:" + alt_env["PATH"]
    sp = subprocess.Popen(cmd_lst, env=alt_env, 
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          shell=shell)
    return(sp.communicate())

def which_ui():
    if 'CDSW_ENGINE_ID'   in os.environ:
        return 'cdsw'
    elif 'JPY_PARENT_PID' in os.environ:
        return 'jupyter'
    elif 'PYCHARM_HOSTED' in os.environ:
        return 'pycharm'
    elif 'RSTUDIO' in os.environ:
        return 'rstudio'
    elif sys.platform == 'linux':
        return 'edgenode'
    elif sys.platform == 'darwin':
        return 'local_mac'
    elif sys.platform in ["win32","msys"]:
        return 'local_win'
    else:
        return 'unknown'
  
def launch_term(cli=''):
    ui = which_ui() 
    if ui in ['cdsw','jupyter']:
        if ui == 'cdsw':
            id=os.environ["CDSW_ENGINE_ID"]
            cd=os.environ["CDSW_DOMAIN"]
            tp = [p for p in sp(["ps","aux"])[0].split() \
                  if b'--path' in p][0][8:]
            url="http://tty-%s.%s/%s/"%(id,cd,tp.decode('utf8'))
            picker=['.','.picker']
        elif ui == 'jupyter':
            # by convention, we want jupyter to run from our project root which has an nb folder:
            url = ['..',]
            picker = []
            path = os.getcwd().split('/')
            while 'nb' not in os.listdir('/'.join(path)):
                url.append('..')
                picker.append('..')
                path = path[:-1]
            url.append('terminals')
            url.append(cli.split(' ')[0].split('-')[0])
            url = '/'.join(url)
            picker.append('.picker')
        else:
            url = 'about:blank'
        with open("/".join(picker),"wb") as f:
            f.write(cli.encode('utf8'))
        script='<script type="text/Javascript">window.open("%s");</script>'%url
        display(HTML(script))
        time.sleep(6) 
        with open("/".join(picker),"wb") as f:
            f.write(''.encode('utf-8'))
    else:
        print('TODO: Write unsupported environment')

def launch_url(url=None):
    """
    Launches url, intended for cdsw and jupyter environments
    """
    script='<script type="text/Javascript">window.open("%s");</script>'%url
    display(HTML(script))
    time.sleep(4)
    
def standard_conn_qry(qry):
    """
    Standard form for connection queries
    
    Allows grouping of commands using ";" seperator
    Each command can have the following form:
       Standard executable statement
       single file, read as sql
    """
    if qry.__class__ == str:
        qry=[q for q in qry.split(";") if q!='']
    rslt=[]
    for q in qry:
        if q[0]=="-":
            q=" ".join(q.split()[1:]).strip('"').strip("'")
        if "." in q.split()[0]:
            f==q.split()[0]
            if "/" not in f:
                os.environ['PROJECT_HOME']="/home/cdsw"
                f=os.environ['PROJECT_HOME']+"/sql/"+f
            f=open(f,'r')
            l=[x for x in f.read().split('\n') if x[:2]!='--']
            q=[re.sub('\\s+',' ',x) for x in " ".join(l).split(";") if x!='']
        rslt+=[q,]
    return(rslt)


def standard_sess_qry(qry):
    """
    Standard form for session queries
    
    Does not allow grouping of commands using ";" seperator
    only first query selected if found.
    Each command can have the following form:
       Standard executable statement
       single file, read as sql
    """
    if qry.__class__ == str:
        qry=qry.split(";")[0]
    q=qry
    if "." in q.split()[0]:
        f==q.split()
        if "/" not in f:
            os.environ['PROJECT_HOME']="/home/cdsw"
            f=os.environ['PROJECT_HOME']+"/sql/"+f
        f=open(f,'r')
        l=[x for x in f.read().split('\n') if x[:2]!='--']
        q=[re.sub('\\s+',' ',x) for x in " ".join(l).split(";") if x!='']
    return(q)

def no_return(qry):
    """
    Query logic to capture queries that have no return
    other than successfull or not"""
    return(("CREATE TABLE" in qry.upper()) |
           (("SELECT" not in qry.upper()) & ("SHOW" not in qry.upper())) )


from setuptools import setup

def readme():
    with open('README.rst', 'r') as f:
        return f.read()

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Environment :: MacOS X',
    'Environment :: Win32 (MS Windows)',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering',
    'Topic :: Database']

setup(name='dsws',
      version='0.17',
      description='Data Science Work Space for Jupyter & CDSW',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='https://pypi.org/project/dsws/',
      author='Brad Barker',
      author_email='brad@ratiocinate.com',
      license='MIT',
      packages=['dsws'],
      classifiers=CLASSIFIERS,
      platforms='any',
      python_requires='>=3.6',
      install_requires=[
          'pandas',
          'numpy'],
          #'thrift==0.9.3'
          #'sasl>=0.2.1'
          #'thrift_sasl==0.2.1'
          #'iopro'
          #'pyodbc'
          #'impyla'
      zip_safe=False)



"""
Workspace connection for pyspark

ws_pyspark is a dsws sess type library that does not follow PEP-249.
It contains a single class, Pyspark.
It is intended to be the primary instantiation for 
insteractive spark sessions.
"""

import pandas as pd
from dsws.util import pretty
from dsws.util import sp
from dsws.util import standard_sess_qry
from os import environ as env
import re
from ast import literal_eval
from pyspark.sql import SparkSession

class Pyspark:

    def __init__(self,kwargs,command=None):
        """
        Workspace class from the pyspark library

        As convntion for DSWS, the following class methods are
        provided:
         - init: returns a session class pyspark.sql.session.SparkSession
         - qry: precesses a query, return type is dependent upon r_type requested:
            - df: pandas dataframe
            - disp: pretty html form of pandas dataframe
            - raw: not yet available, to return spark dataframe, pyspark.sql.dataframe.DataFrame
            - cmd: not yet available standard query form to be processed. Helpful for debugging and logging.
        
        Configuring a Pyspark class is typically done through dsws.duct, but can also be
        evaluated as:
        ```python
        from dsws.ws_pyspark import Pypsark
        kwargs = {'spark.yarn.executor.memoryOverhead':2048,
                  'spark.sql.shuffle.partitions':10,
                  'spark.yarn.queue':'root.default',
                  'spark.app.name':'dsws_session'}
        spark = Pyspark(kwargs)
        ```
        
        To know which values can be accepted in kwargs run, goto Apache Spark
        documentation for your respecitive version:
        https://spark.apache.org/docs/latest/configuration.html
        
        spark-sql is the most logical cli to use for the spark command. However, distributions
        may not include spark-sql and therefore will not be in bin path as expected. Simply,
        do not populate the command argument.
        https://spark.apache.org/docs/latest/sql-distributed-sql-engine.html#running-the-spark-sql-cli
        """
        self.conf=kwargs
        if command is not None:
            self.command=command

    def init(self):
        ss=SparkSession.builder
        for k,v in self.conf.items():
            ss=ss.config(k,v)
        return(ss.getOrCreate())
        
    def qry(self,qry,r_type="df",limit=20):
        qry=standard_sess_qry(qry)
        spark=self.init()
        if r_type=="disp" and "LIMIT" not in qry.upper() and "SELECT" in qry.upper():
            qry = qry + " LIMIT " + str(limit)
        rslt=spark.sql(qry).toPandas()
        if r_type == "disp":
            pretty(rslt,col="#F4A460")
            rslt=None
        return(rslt)



"""
Workspace connection for Ibis_impala

It contains a single class, Ibis_impala.
While Ibis impala presumes that the ibis user will
 be using impala with HDFS.
"""

import ibis
import pandas          as pd
from dsws.util     import pretty
from dsws.util     import sp
from dsws.util     import standard_conn_qry
from dsws.util     import no_return
from os import environ as env
from ast           import literal_eval


class Ibis_impala:

    def __init__(self,kwargs,command=None):
        """
        Workspace class from the ibis library

        Similar for other DSWS classes, the following class methods are
        provided:
         - hdfs: returns an hdfs conneciton
         - impalaClient: returns an ibis impala client 
         - qry: precesses a query, return type is dependent upon r_type requested:
            - df/raw: pandas dataframe
            - disp: pretty html form of pandas dataframe
            - msg: experimental - retuens query profile if using impala
            - cmd: standard query form to be processed. Helpful for debugging and logging
        
        Configuring an ibis class is typically done through dsws.duct, but can also be
        evaluated as:
        ```python
        from dsws.ws_ibis_impala import Ibis_impala
        kwargs = {'ibis.config':
                  {'impala.temp_db':'default',
                   'impala.temp_hdfs_path':'/usr/hive/warehouse/default'},
                  'hdfs.connect.kwargs':
                  {'host':'webhdfs.ratiocinate.com',
                   'port':14000,
                   'protocol':'webhdfs',
                   'auth_mechanism':'GSSAPI',
                   'use_https':True,
                   'verify':False},
                  'impala.connect.kwargs':
                  {'host':'impala',
                   'port':21050,
                   'database':'default',
                   'use_ssl':True,
                   'auth_mechanism':'GSSAPI'},
                  'impala.options':
                  {'request_pool':'"root.default"'}}

        ib = Ibis_impala(kwargs)
        ```
        Contrary to other common connections, Ibis for impala  requires both an hdfs and impala 
        connnections in addition to ibis config and impala options for setting request_pool"""
 
        self.ibis_config=kwargs['ibis.config']
        self._set_ibis_config()
        self.hdfs_connect_kwargs=kwargs['hdfs.connect.kwargs']
        self.impala_connect_kwargs=kwargs['impala.connect.kwargs']
        self.impala_options=kwargs['impala.options']
        if command is not None:
            self.command=command
            
    def _set_ibis_config(self):
        for k in self.ibis_config.keys():
            ibis.config.set_option(k,self.ibis_config[k])
            
    def hdfs(self):
        return ibis.hdfs_connect(**self.hdfs_connect_kwargs)
    
    def impalaClient(self):
        kwargs=self.impala_connect_kwargs
        kwargs.update({'hdfs_client':self.hdfs()})
        client=ibis.impala.connect(**kwargs)
        client.set_options(self.impala_options)
        return client
    
    def qry(self,qry,r_type="df",limit=30):
        r_type = 'df' if r_type=='raw' else r_type
        qry=standard_conn_qry(qry)
        if r_type=="cmd":
            return(qry)
        client = self.impalaClient()
        for q in qry[:-1]:
            client.raw_sql(q,False)
        if r_type=="disp" and "LIMIT" not in qry[-1].split()[-2].upper() and "SELECT" in qry[-1].split()[1].upper():
            qry[-1] = qry[-1] + " LIMIT " + str(limit)
        if no_return(qry[-1]):
            return(None)
        if r_type in ("df","disp"):
            rslt = client.sql(qry[-1]).execute()
            if r_type=="disp":
                pretty(rslt,col="#E07862")
                rslt=None
        elif r_type=="msg":
            #TODO: Add messaging capability
            rslt=None
        else:
            rslt=None
        client.close()
        return(rslt)
